def element_node(doc:object, tag:str, attr:str) -> None:
    
    ele = doc.createElement(tag)
    
    if attr is not None:
        text = doc.createTextNode(attr)
        ele.appendChild(text)
    
    return ele

def child_node(doc:object, tag:str, attr:str, parent_node:object) -> None:
    
    child = element_node(doc,tag,attr)
    parent_node.appendChild(child)

def object_node(doc:object, attr:list|tuple, box:list|tuple, seg:list[list,list]=None) -> object:
    
    object = doc.createElement("object")
    child_node(doc,"name",box[0],object)
    child_node(doc,"pose",str(attr[0]),object)
    child_node(doc,"truncated",str(attr[1]),object)
    child_node(doc,"difficult",str(attr[2]),object)
    child_node(doc,"occluded",str(attr[3]),object)

    bnd = doc.createElement("bndbox")
    child_node(doc,"xmin",str(box[1]),bnd)
    child_node(doc,"xmax",str(box[3]),bnd)
    child_node(doc,"ymin",str(box[2]),bnd)
    child_node(doc,"ymax",str(box[4]),bnd)
    object.appendChild(bnd)
    
    if seg is not None:

        poly = doc.createElement("polygon")
        
        for idx, (x, y) in enumerate(zip(*seg)):
            child_node(doc,"x"+str(idx+1),str(x),poly)
            child_node(doc,"y"+str(idx+1),str(y),poly)
        object.appendChild(poly)

    return object


import xml.etree.ElementTree as et
import tqdm
import cv2
import os

def class_getter(label_path:str) -> list:

    cll = []
    len_ann = []

    for file in os.listdir(label_path):
        if file.endswith(".xml"):
            tree = et.parse(os.path.join(label_path,file))
            
            for child in tree.findall("object"):
                cll.append(child.find("name").text)
        
        elif file.endswith(".txt"):

            with open(os.path.join(label_path,file),"r") as f:
                for sep in f.read().split("\n"):
                    len_ann.append(len(sep.split()))
                    cll.append(sep.split()[0])

    if len(set(len_ann)) == 1 and list(set(len_ann))[0] == 5:
        print("\033[1m\033[4mThe lables are annotated for Detection task.\033[0m\n")
    
    elif len(set(len_ann)) > 1:
        print("\033[1m\033[4mThe lables are annotated for Segmentation task.\033[0m\n")  
    
    return sorted(list(set(cll)))




def coco_validator(coco:object, image_path:str) -> None:
    img_dt = coco.loadImgs(coco.getImgIds())
    ann_dt = coco.loadAnns(coco.getAnnIds())
    cat_dt = coco.loadCats(coco.getCatIds())
    assert len(img_dt) == len(os.listdir(image_path)), "Number of images does not match."
    
    filefmt = [i["file_name"][-4:] for i in img_dt]
    imgfmt = [i[-4:] for i in os.listdir(image_path)]
    assert len(set(filefmt)) == 1 or len(set(imgfmt)) == 1, "All images must be in a single format."
    assert set(filefmt).issubset([".jpg",".png"]) and set(imgfmt).issubset([".jpg",".png"]), "Images must be in jpg or png format."

    filename = [i["file_name"] for i in img_dt]
    assert sorted(filename) == sorted(os.listdir(image_path)), "Image files does not match."

    catinAnn = [i["category_id"] for i in ann_dt]
    assert len(set(catinAnn)) == len(cat_dt), "Categories does not match."

    imginAnn = [i["image_id"] for i in ann_dt]
    assert len(set(imginAnn)) == len(img_dt), "Image id does not match."
    
    print("\nChecking width and height........\n")
    for i in tqdm.tqdm(img_dt):
        img = cv2.imread(os.path.join(image_path,i["file_name"]))
        assert (i["height"], i["width"]) == img.shape[:2], "Image shape does not match."
    
    print("\nChecking data in annotations.......\n")
    for i in tqdm.tqdm(ann_dt):
        assert i["area"]>0, "Area values must be greater than 0."
        assert len(i["bbox"]) % 2 == 0 and len(i["bbox"]) == 4, "Invalid bbox values."
        t_img_dt = coco.loadImgs(i["image_id"])

        if i["segmentation"] != [[]]:
            if isinstance(i["segmentation"],dict):
                assert sorted(list(i["segmentation"].keys())) == sorted(["counts","size"]), "RLE Segmentation is incorrect"
                assert len(i["segmentation"]["size"]) == 2 , "RLE size values are corrupted"
                assert sum(i["segmentation"]["counts"]) == i["segmentation"]["size"][0] * i["segmentation"]["size"][1], "RLE counts values are corrupted."
                continue

            assert len(i["segmentation"][0]) % 2 == 0 and len(i["segmentation"][0]) >= 6, "Invalid segemntation values."
            

            seg_x = i["segmentation"][0][::2]
            seg_y = i["segmentation"][0][1::2]
            

            for x,y in zip(seg_x,seg_y):
                assert x <= t_img_dt[0]["width"], "Segmentation values are corrupted."
                assert y <= t_img_dt[0]["height"], "Segmentation values are corrupted."
        
        for x,y in zip(i["bbox"][::2],i["bbox"][1::2]):
            assert x <= t_img_dt[0]["width"], "Bbox values are corrupted."
            assert y <= t_img_dt[0]["height"], "Bbox values are corrupted."

    print("\nAnnotation Validator is complete.")
    return 




def voc_validator(image_path:str, labels_path:str) -> list:

    assert len(os.listdir(image_path)) == len(os.listdir(labels_path)), "Images and Labels does not match."
    assert sorted([i[:-4] for i in os.listdir(image_path)]) == sorted([i[:-4] for i in os.listdir(labels_path)]), "Images and labels files does not match."

    imgfmt = [i[-4:] for i in os.listdir(image_path)]
    labelfmt = [i[-4:] for i in os.listdir(labels_path)]
    assert len(set(labelfmt)) == 1 and list(set(labelfmt))[0] == ".xml" , "Labels must be in xml format."
    assert len(set(labelfmt)) == 1 or len(set(imgfmt)) == 1, "All images and labels must be in a single format."
    assert set(imgfmt).issubset([".jpg",".png"]), "Images must be in jpg or png format."

    print("\nChecking annotation files.........\n")

    cls = class_getter(labels_path)

    print("Classes present in all the labels : {}\n".format(cls))

    for file in tqdm.tqdm(os.listdir(image_path)):
        tree = et.parse(os.path.join(labels_path,file[:-4]+".xml"))
        filename = tree.find("filename").text
        width = int(tree.find("size").find("width").text)
        height = int(tree.find("size").find("height").text)
        
        assert file == filename, "Filename does not match."

        img = cv2.imread(os.path.join(image_path,filename))
        assert (height, width) == img.shape[:2], "Image shape does not match."

        for ann in tree.findall("object"):
            bbox = ann.find("bndbox")
            assert len(bbox) % 2 == 0 and len(bbox) == 4, "Invalid bbox values."

            xmin = int(bbox.find("xmin").text)
            ymin = int(bbox.find("ymin").text)
            xmax = int(bbox.find("xmax").text)
            ymax = int(bbox.find("ymax").text)

            for x,y in zip([xmin,xmax],[ymin,ymax]):
                assert x <= width, "Bbox values are corrupted."
                assert y <= height, "Bbox values are corrupted."
            
            if ann.find("polygon") is not None:
                seg = [float(child.text) for child in ann.find("polygon")]
                assert len(seg) % 2 == 0 and len(seg) >= 6, "Invalid segemntation values."

                seg_x = seg[::2]
                seg_y = seg[1::2]

                for x,y in zip(seg_x,seg_y):
                    assert x <= width, "Segmentation values are corrupted."
                    assert y <= height, "Segmentation values are corrupted."

    print("\nAnnotation Validator is complete.")
    return cls




def yolo_validator(image_path:str, labels_path:str, n_classes:list[int]) -> None:

    assert len(os.listdir(image_path)) == len(os.listdir(labels_path)), "Images and Labels does not match."
    assert sorted([i[:-4] for i in os.listdir(image_path)]) == sorted([i[:-4] for i in os.listdir(labels_path)]), "Images and labels files does not match."

    imgfmt = [i[-4:] for i in os.listdir(image_path)]
    labelfmt = [i[-4:] for i in os.listdir(labels_path)]
    assert len(set(labelfmt)) == 1 and list(set(labelfmt))[0] == ".txt" , "Labels must be in txt format."
    assert len(set(labelfmt)) == 1 or len(set(imgfmt)) == 1, "All images and labels must be in a single format."
    assert set(imgfmt).issubset([".jpg",".png"]), "Images must be in jpg or png format."

    print("\nChecking annotation files.........\n")

    cls = class_getter(labels_path)

    assert len(n_classes) == len(cls), "Classes does not match."

    print("There are {} classes present in all the labels.\n".format(len(cls)))

    for file in tqdm.tqdm(os.listdir(image_path)):
         img = cv2.imread(os.path.join(image_path,file))
         height, width = img.shape[:2]
         
         with open(os.path.join(labels_path,file[:-4]+".txt"),"r") as f:
                for sep in f.read().split("\n"):
                    data = sep.split()[1:]

                    for i in data:
                        assert float(i) <= 1, "Invalid label content."
                    
                    if len(data) > 4:
                        assert len(data) % 2 == 0 and len(data) >= 6, "Invalid segemntation values."
                        
                        seg = [round(float(val)*width,2) if idx%2 == 0 else round(float(val)*height,2) for idx,val in enumerate(data)]
                        seg_x = seg[::2]
                        seg_y = seg[1::2]

                        for x,y in zip(seg_x,seg_y):
                            assert x <= width, "Segmentation values are corrupted."
                            assert y <= height, "Segmentation values are corrupted."
                    
                    elif len(data) == 4:
                        assert len(data) % 2 == 0, "Invalid bbox values."

                        w, h = float(data[2])*width, float(data[3])*height
                        x, y = round(((float(data[0])*2*width)-w)/2,2), round(((float(data[1])*2*height)-h)/2,2)

                        for x,y in zip([x,x+w],[y,y+h]):
                            assert x <= width, "Bbox values are corrupted."
                            assert y <= height, "Bbox values are corrupted."
    
    print("\nAnnotation Validator is complete.")
    return 


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from torch import mean


ALPHA_focal = 0.8
GAMMA_focal = 2

ALPHA_tversky = 0.5
BETA_tversky = 0.5

ALPHA_focal_tversky = 0.5
BETA_focal_tversky = 0.5
GAMMA_focal_tversky = 1

ALPHA_combo = 0.5 
BETA_combo = 0.5
CE_RATIO_combo = 0.5

def flatten_probas(input, target, ignore_index):
    num_class = input.size(1)
    input = input.permute(0, 2, 3, 1).contiguous()
    
    input_flatten = input.view(-1, num_class)
    target_flatten = target.view(-1)
    
    mask = (target_flatten != ignore_index)
    input_flatten = input_flatten[mask]
    target_flatten = target_flatten[mask]
    
    return input_flatten, target_flatten

def flatten_binary_scores(scores, labels, ignore=None):
    """
    Flattens predictions in the batch (binary case)
    Remove labels equal to 'ignore'
    """
    scores = scores.view(-1)
    labels = labels.view(-1)
    if ignore is None:
        return scores, labels
    valid = (labels != ignore)
    vscores = scores[valid]
    vlabels = labels[valid]
    return vscores, vlabels

def lovasz_grad(gt_sorted):
    """
    Computes gradient of the Lovasz extension w.r.t sorted errors
    """
    p = len(gt_sorted)
    gts = gt_sorted.sum()
    intersection = gts - gt_sorted.float().cumsum(0)
    union = gts + (1 - gt_sorted).float().cumsum(0)
    jaccard = 1. - intersection / union
    if p > 1:
        jaccard[1:p] = jaccard[1:p] - jaccard[0:-1]
    return jaccard

def lovasz_hinge(logits, labels, per_image=True, ignore=None):
    """
    Binary Lovasz hinge loss
      logits: [B, H, W] Variable, logits at each pixel (between -\infty and +\infty)
      labels: [B, H, W] Tensor, binary ground truth masks (0 or 1)
      per_image: compute the loss per image instead of per batch
      ignore: void class id
    """
    if per_image:
        loss = mean(lovasz_hinge_flat(*flatten_binary_scores(log.unsqueeze(0), lab.unsqueeze(0), ignore))
                          for log, lab in zip(logits, labels))
    else:
        loss = lovasz_hinge_flat(*flatten_binary_scores(logits, labels, ignore))
    return loss

def lovasz_hinge_flat(logits, labels):
    """
    Binary Lovasz hinge loss
      logits: [P] Variable, logits at each prediction (between -\infty and +\infty)
      labels: [P] Tensor, binary ground truth labels (0 or 1)
      ignore: label to ignore
    """
    if len(labels) == 0:
        return logits.sum() * 0.
    signs = 2. * labels.float() - 1.
    errors = (1. - logits * Variable(signs))
    errors_sorted, perm = torch.sort(errors, dim=0, descending=True)
    perm = perm.data
    gt_sorted = labels[perm]
    grad = lovasz_grad(gt_sorted)
    loss = torch.dot(F.relu(errors_sorted), Variable(grad))
    return loss


def lovasz_softmax(probas, labels, classes='present', per_image=False, ignore=None):
    """
    Multi-class Lovasz-Softmax loss
      probas: [B, C, H, W] Variable, class probabilities at each prediction (between 0 and 1).
              Interpreted as binary (sigmoid) output with outputs of size [B, H, W].
      labels: [B, H, W] Tensor, ground truth labels (between 0 and C - 1)
      classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
      per_image: compute the loss per image instead of per batch
      ignore: void class labels
    """
    if per_image:
        loss = mean(lovasz_softmax_flat(*flatten_probas(prob.unsqueeze(0), lab.unsqueeze(0), ignore), classes=classes)
                          for prob, lab in zip(probas, labels))
    else:
        loss = lovasz_softmax_flat(*flatten_probas(probas, labels, ignore), classes=classes)
    return loss


def lovasz_softmax_flat(probas, labels, classes='present'):
    """
    Multi-class Lovasz-Softmax loss
      probas: [P, C] Variable, class probabilities at each prediction (between 0 and 1)
      labels: [P] Tensor, ground truth labels (between 0 and C - 1)
      classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
    """
    if probas.numel() == 0:
        return probas * 0.
    C = probas.size(1)
    losses = []
    class_to_sum = list(range(C)) if classes in ['all', 'present'] else classes
    for c in class_to_sum:
        fg = (labels == c).float()
        if (classes is 'present' and fg.sum() == 0):
            continue
        if C == 1:
            if len(classes) > 1:
                raise ValueError('Sigmoid output possible only with 1 class')
            class_pred = probas[:, 0]
        else:
            class_pred = probas[:, c]
        errors = (Variable(fg) - class_pred).abs()
        errors_sorted, perm = torch.sort(errors, 0, descending=True)
        perm = perm.data
        fg_sorted = fg[perm]
        losses.append(torch.dot(errors_sorted, Variable(lovasz_grad(fg_sorted))))
    return mean(losses)


class DiceLoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(DiceLoss, self).__init__()

    def forward(self, inputs, targets, smooth=1, sigmoid=False):
        
        if sigmoid:
            inputs = F.sigmoid(inputs)       
        
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        
        intersection = (inputs * targets).sum()                            
        dice = (2.*intersection + smooth)/(inputs.sum() + targets.sum() + smooth)  
        
        return 1 - dice


class DiceBCELoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(DiceBCELoss, self).__init__()

    def forward(self, inputs, targets, smooth=1, sigmoid=False):
        
        if sigmoid:
            inputs = F.sigmoid(inputs)        
        
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        
        intersection = (inputs * targets).sum()                            
        dice_loss = 1 - (2.*intersection + smooth)/(inputs.sum() + targets.sum() + smooth)  
        BCE = F.binary_cross_entropy(inputs, targets, reduction='mean')
        Dice_BCE = BCE + dice_loss
        
        return Dice_BCE
    

class IoULoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(IoULoss, self).__init__()

    def forward(self, inputs, targets, smooth=1, sigmoid=False):
        
        if sigmoid:
            inputs = F.sigmoid(inputs)       
        
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        
        intersection = (inputs * targets).sum()
        total = (inputs + targets).sum()
        union = total - intersection 
        
        IoU = (intersection + smooth)/(union + smooth)
                
        return 1 - IoU
    

class FocalLoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(FocalLoss, self).__init__()

    def forward(self, inputs, targets, alpha=ALPHA_focal, gamma=GAMMA_focal, smooth=1, sigmoid=False):
        
        if sigmoid:
            inputs = F.sigmoid(inputs)       
        
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        
        BCE = F.binary_cross_entropy(inputs, targets, reduction='mean')
        BCE_EXP = torch.exp(-BCE)
        focal_loss = alpha * (1-BCE_EXP)**gamma * BCE
                       
        return focal_loss
    


class TverskyLoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(TverskyLoss, self).__init__()

    def forward(self, inputs, targets, smooth=1, alpha=ALPHA_tversky, beta=BETA_tversky, sigmoid=False):
        
        if sigmoid:
            inputs = F.sigmoid(inputs)       
        
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        
        TP = (inputs * targets).sum()    
        FP = ((1-targets) * inputs).sum()
        FN = (targets * (1-inputs)).sum()
       
        Tversky = (TP + smooth) / (TP + alpha*FP + beta*FN + smooth)  
        
        return 1 - Tversky
    


class FocalTverskyLoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(FocalTverskyLoss, self).__init__()

    def forward(self, inputs, targets, smooth=1, alpha=ALPHA_focal_tversky, beta=BETA_focal_tversky, gamma=GAMMA_focal_tversky, sigmoid=False):
        
        if sigmoid:
            inputs = F.sigmoid(inputs)       
        
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        
        TP = (inputs * targets).sum()    
        FP = ((1-targets) * inputs).sum()
        FN = (targets * (1-inputs)).sum()
        
        Tversky = (TP + smooth) / (TP + alpha*FP + beta*FN + smooth)  
        FocalTversky = (1 - Tversky)**gamma
                       
        return FocalTversky



class LovaszHingeLoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(LovaszHingeLoss, self).__init__()

    def forward(self, inputs, targets,sigmoid=False):
        if sigmoid:
            inputs = F.sigmoid(inputs)    
        Lovasz = lovasz_hinge(inputs, targets, per_image=False)                       
        return Lovasz
    

class ComboLoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(ComboLoss, self).__init__()

    def forward(self, inputs, targets, smooth=1, alpha=ALPHA_combo, beta=BETA_combo, eps=1e-9):
        
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        
        intersection = (inputs * targets).sum()    
        dice = (2. * intersection + smooth) / (inputs.sum() + targets.sum() + smooth)
        
        inputs = torch.clamp(inputs, eps, 1.0 - eps)       
        out = - (ALPHA_combo * ((targets * torch.log(inputs)) + ((1 - ALPHA_combo) * (1.0 - targets) * torch.log(1.0 - inputs))))
        weighted_ce = out.mean(-1)
        combo = (CE_RATIO_combo * weighted_ce) - ((1 - CE_RATIO_combo) * dice)
        
        return combo


import os
import cv2
import tqdm
import imageio as iio
from dateutil import relativedelta




def vid2img(video_path:str, destination_path:str, size:tuple[int,int]) -> None:

    assert os.path.isfile(video_path), "video_path must be a path to a video file."

    vid = cv2.VideoCapture(video_path)

    total_length = int(vid.get(cv2.CAP_PROP_FRAME_COUNT))
    vid_fps = int(vid.get(cv2.CAP_PROP_FPS))
    vid_time = relativedelta.relativedelta(seconds=(total_length//vid_fps))

    print("\nTotal time of the video : [ {}:{}:{} ]\n".format(vid_time.hours, vid_time.minutes, vid_time.seconds))

    for i in tqdm.tqdm(range(total_length)):
        ret, frame = vid.read()
        if ret:
            if size is not None:
                frame = cv2.resize(frame, size)

            cv2.imwrite(os.path.join(destination_path, f"frame-{i+1}.png"), frame)
    
    vid.release()
    
    print("\nExtraction Successfull.")

    print(f"\nA Total of \033[1m\033[4m{total_length} images\033[0m has been extracted.")

    print("\nThe Extracted images are saved at '{}'.".format(destination_path))




def fps_changer(video_path:str, fps:int, destination_path:str, filename:str,
                size:tuple[int,int], is_fast:bool) -> None:

    assert os.path.isfile(video_path), "video_path must be a path to a video file."

    vid = cv2.VideoCapture(video_path)

    if size is None:
        size = (int(vid.get(cv2.CAP_PROP_FRAME_WIDTH)), int(vid.get(cv2.CAP_PROP_FRAME_HEIGHT)))
        
    if filename is not None:
        assert filename.endswith(".mp4"), "filename must ends with '.mp4'."
    else:
        filename = video_path.split("/")[-1][:-4]+"-fps_changed.mp4"
    
    total_length = int(vid.get(cv2.CAP_PROP_FRAME_COUNT))
    vid_fps = int(vid.get(cv2.CAP_PROP_FPS))
    vid_time = relativedelta.relativedelta(seconds=(total_length//vid_fps))

    if fps is None:
        fps = vid_fps

    print(f"\nThe video FPS : \033[1m\033[4m{vid_fps} fps\033[0m .")

    assert (not is_fast and fps > vid_fps) or (is_fast and fps <= vid_fps and fps > 0), "'is_fast' must be set to false if 'fps' is greater than video fps."

    vid_wr = cv2.VideoWriter(os.path.join(destination_path, filename), cv2.VideoWriter.fourcc(*"mp4v"), fps, size)

    step = vid_fps//fps

    if is_fast and step > 1:  
        spots = list(range(0,total_length,step))

    print("\nTotal time of the video : [ {}:{}:{} ]\n".format(vid_time.hours, vid_time.minutes, vid_time.seconds))

    for i in tqdm.tqdm(range(total_length)):

        ret, frame = vid.read()
        if ret:
            if size is not None:
                frame = cv2.resize(frame, size)
            
            if is_fast and step > 1:
                if i in spots:
                    vid_wr.write(frame)
                continue
            
            vid_wr.write(frame)
    
    vid_wr.release()
    vid.release()

    print("\nConversion Successfull.")

    print(f"\nA Total of \033[1m\033[4m{total_length//step if is_fast and step > 1 else 0} images\033[0m has been dropped from the video.")

    print("\nThe Converted video is saved at '{}' named '{}'.".format(destination_path, filename))




def vid_cropper(video_path:str, start_time:str, end_time:str, destination_path:str,
                filename:str, fps:int, size:tuple[int,int]) -> None:

    assert os.path.isfile(video_path), "video_path must be a path to a video file."
    assert len(start_time.strip().split(":")) == 3 and len(end_time.strip().split(":")) == 3, "start_time and end_time must be in h:m:s format."

    st_hr, st_mnt, st_sec= start_time.strip().split(":")
    ed_hr, ed_mnt, ed_sec= end_time.strip().split(":")
    test = st_hr+st_mnt+st_sec+ed_hr+ed_mnt+ed_sec
    
    assert test.isdigit(), "start_time and end_time must be in digits."
    assert int(st_mnt) < 60 and int(ed_mnt) < 60 and int(st_sec) < 60 and int(ed_sec) < 60 and int(st_hr) < 24 and int(ed_hr) < 24, "start_time and end_time must be in time format."

    start_second = sum([int(st_hr)*3600, int(st_mnt)*60, int(st_sec)])
    end_second = sum([int(ed_hr)*3600, int(ed_mnt)*60, int(ed_sec)])

    assert end_second > start_second, "start_time must be lower than end_time."

    vid = cv2.VideoCapture(video_path)

    if size is None:
        size = (int(vid.get(cv2.CAP_PROP_FRAME_WIDTH)), int(vid.get(cv2.CAP_PROP_FRAME_HEIGHT)))
    
    if filename is not None:
        assert filename.endswith(".mp4"), "filename must ends with '.mp4'."
    else:
        filename = video_path.split("/")[-1][:-4]+"-cropped.mp4"
    
    total_length = int(vid.get(cv2.CAP_PROP_FRAME_COUNT))
    vid_fps = int(vid.get(cv2.CAP_PROP_FPS))
    tot_sec = (total_length//vid_fps)
    vid_time = relativedelta.relativedelta(seconds=tot_sec)
    
    if fps is None:
        fps = vid_fps
    
    crp_vid_time = relativedelta.relativedelta(seconds=(end_second - start_second))

    print("\nTotal time of the video : [ {}:{}:{} ]".format(vid_time.hours, vid_time.minutes, vid_time.seconds))

    print(f"\nThe video FPS : \033[1m\033[4m{vid_fps} fps\033[0m .\n")

    assert end_second <= tot_sec and start_second < tot_sec, "The specified time exceeds the video time range."
    assert fps <= vid_fps and fps > 0, "fps must not exceed the video fps"

    vid_wr = cv2.VideoWriter(os.path.join(destination_path, filename), cv2.VideoWriter.fourcc(*"mp4v"), fps, size)

    start_frame = start_second*vid_fps
    end_frame = end_second*vid_fps

    for i in tqdm.tqdm(range(total_length)):
        
        ret, frame = vid.read()
        
        if i >= start_frame and i <= end_frame and ret:
            if size is not None:
                frame = cv2.resize(frame, size)

            vid_wr.write(frame)
        
        elif i == end_frame:
            break
    
    vid_wr.release()
    vid.release()

    print("\nCropping Successfull.")

    print("\nA Total duration of cropped video : [ {}:{}:{} ]".format(crp_vid_time.hours, crp_vid_time.minutes, crp_vid_time.seconds))

    print("\nThe Cropped video is saved at '{}' named '{}'.".format(destination_path, filename))

            


def vid2gif(video_path:str, destination_path:str, filename:str, size:tuple[int,int],
            duration_of_frame:int, loop:int) -> None:
    
    print(size)
    assert os.path.isfile(video_path), "video_path must be a path to a video file."

    if filename is not None:
        assert filename.endswith(".gif"), "filename must ends with '.gif'."
    else:
        filename = video_path.split("/")[-1][:-4]+".gif"

    vid = cv2.VideoCapture(video_path)

    if size is None:
        size = (int(vid.get(cv2.CAP_PROP_FRAME_WIDTH)), int(vid.get(cv2.CAP_PROP_FRAME_HEIGHT)))

    total_length = int(vid.get(cv2.CAP_PROP_FRAME_COUNT))
    vid_fps = int(vid.get(cv2.CAP_PROP_FPS))
    tot_sec = (total_length//vid_fps)
    vid_time = relativedelta.relativedelta(seconds=tot_sec)

    print("\nTotal time of the video : [ {}:{}:{} ]\n".format(vid_time.hours, vid_time.minutes, vid_time.seconds))

    with iio.get_writer(os.path.join(destination_path,filename), mode="I", duration=duration_of_frame, loop=loop) as writer:

        for _ in tqdm.tqdm(range(total_length)):

            ret, frame = vid.read()
            if ret:
                if size is not None:
                    frame = cv2.resize(frame, size)

                writer.append_data(cv2.cvtColor(frame,cv2.COLOR_BGR2RGB))
        
    
    print("\nConversion Completed.\n")

    print("The gif file is saved at '{}' named '{}'.".format(destination_path, filename))


import xml.etree.ElementTree as et
import os
import tqdm

from helpers_ai.utils.validator import voc_validator, yolo_validator, class_getter




def __check_corrupt(corrupted):

    if len(corrupted)>0:

        print("\nThere are {} corrupted files.\n".format(len(corrupted)))

        for i in corrupted:
            print(i,"data is corrupted.")
        
        print("")
        
        raise ValueError("Corrupted files are present")




def detection(label_path:str, image_path:str, destination:str) -> None:

    print("\nAnnotation Validator in progress.....")

    voc_validator(image_path,label_path)

    corrupted = []
    s_cls = class_getter(label_path)

    print("\nThere are {} classes are present in the given labels_path.".format(len(s_cls)))
    flg = input("\nContinue[y/n]: ")
    print("")

    if flg == "y" or flg =="Y":
    
        for file in tqdm.tqdm(os.listdir(label_path)):
            if file.endswith(".xml"):
                tree = et.parse(os.path.join(label_path,file))
                size = tree.find("size")
                width = int(size.find("width").text)
                height = int(size.find("height").text)
                out = []
                ps = True
                    
                for idx,ann in enumerate(tree.findall("object")):
                    classes = ann.find("name").text
                    bbox = ann.find("bndbox")
                    xmin = int(bbox.find("xmin").text)
                    ymin = int(bbox.find("ymin").text)
                    xmax = int(bbox.find("xmax").text)
                    ymax = int(bbox.find("ymax").text)
                    cx, cy, nw, nh =  (xmin+xmax)/(2*width), (ymin+ymax)/(2*height), (xmax-xmin)/width, (ymax-ymin)/height

                    if cx>1 or cy>1 or nw>1 or nh>1 :
                        corrupted.append(file[:-4]+".txt")
                        ps = False
                        break

                    out.append(f"{s_cls.index(classes)} {cx} {cy} {nw} {nh}")

                if ps:
                    with open(os.path.join(destination,file[:-4]+".txt"),"w+") as f:
                        f.writelines("\n".join(out))

        __check_corrupt(corrupted)
        
        print("\nThe conversion is completed.")

        print("\nValidation of converted annotation in progress..........\n")

        yolo_validator(image_path,destination,s_cls)

        with open(os.path.join(destination,"__classes.txt"),"w+") as f:
            f.writelines("\n".join(s_cls))
        
        print("\nThe annotated file is saved at '{}'.\n".format(destination))




def segment(label_path:str, image_path:str, destination:str) -> None:

    print("\nAnnotation Validator in progress.....")

    voc_validator(image_path,label_path)

    corrupted = []
    s_cls = class_getter(label_path)

    print("\nThere are {} classes are present in the given labels_path.".format(len(s_cls)))
    flg = input("\nContinue[y/n]: ")
    print("")

    if flg == "y" or flg =="Y":

        for file in tqdm.tqdm(os.listdir(label_path)):
            if file.endswith(".xml"):
                tree = et.parse(os.path.join(label_path,file))
                size = tree.find("size")
                width = int(size.find("width").text)
                height = int(size.find("height").text)
                out = []
                ps =True
                
                for idx,ann in enumerate(tree.findall("object")):
                    classes = ann.find("name").text
                    seg = [float(child.text)/width if idx%2 == 0 else float(child.text)/height for idx,child in enumerate(ann.find("polygon"))]
            
                    if max(seg)>1:
                        corrupted.append(file[:-4]+".txt")
                        ps = False
                        break
                    
                    out.append(f"{s_cls.index(classes)} {' '.join([str(i) for i in seg])}")

                if ps:
                    with open(os.path.join(destination,file[:-4]+".txt"),"w+") as f:
                        f.writelines("\n".join(out))

        __check_corrupt(corrupted)

        print("\nThe conversion is completed.")

        print("\nValidation of converted annotation in progress..........\n")

        yolo_validator(image_path,destination,s_cls)

        with open(os.path.join(destination,"__classes.txt"),"w+") as f:
            f.writelines("\n".join(s_cls))
        
        print("\nThe annotated file is saved at '{}'.\n".format(destination))


from xml.dom import minidom
import os
import cv2
import tqdm

from helpers_ai.utils.validator import yolo_validator, voc_validator
from helpers_ai.utils.voc import element_node, child_node, object_node




def convert(labels_path:str, image_path:str, destination:str, n_classes:list[str]) -> None:

    print("\nAnnotation Validator in progress.....")

    yolo_validator(image_path, labels_path, n_classes)

    cl = []
    corrupted =[]

    for file in os.listdir(labels_path):
        
        with open(os.path.join(labels_path,file),"r") as lr:
            raw = lr.read()
            data = [sep.split(" ") for sep in raw.split("\n")]
        
        for ann in data:
            cl.append(ann[0])
            
            if (len(ann)==5) and (float(ann[1])>1 or float(ann[2])>1 or float(ann[3])>1 or float(ann[4])>1):
                corrupted.append(file)
                break

            elif len(ann)>5 and max([float(i) for i in ann[1:]])>1:
                corrupted.append(file)
                break
    
    if len(corrupted)>0:

        print("\nThere are {} corrupted files.\n".format(len(corrupted)))
    
        for i in corrupted:
            print(i,"data is corrupted.")
        
        print("")
        raise ValueError("Corrupted files are present")
        
    if len(set(cl)) != len(n_classes):
        raise AssertionError("The number of classes given does not match the classes in annotated file")
    
    print("")
    
    for file in tqdm.tqdm(os.listdir(image_path)):
        
        img = cv2.imread(os.path.join(image_path,file))
        height,width,depth = img.shape

        with open(os.path.join(labels_path,file[:-4]+".txt"),"r") as f:
            raw = f.read()
            data = [sep.split(" ") for sep in raw.split("\n")]
        
        root = minidom.Document()
        xml = root.createElement("annotation")
        root.appendChild(xml)

        child_node(root,"folder","",xml)
        child_node(root,"filename",file,xml)
        child_node(root,"path",file,xml)

        source = element_node(root,"source",None)
        xml.appendChild(source)
        child_node(root,"database","",source)

        size = element_node(root,"size",None)
        xml.appendChild(size)
        child_node(root,"width",str(width),size)
        child_node(root,"height",str(height),size)
        child_node(root,"depth",str(depth),size)

        child_node(root,"segmented","0",xml)

        if len(data[0])==5:

            for ann in data:
                classes, cx, cy, nw, nh = ann
                w, h = float(nw)*width, float(nh)*height
                x, y = int(((float(cx)*2*width)-w)/2), int(((float(cy)*2*height)-h)/2)
                attr = ["Unspecified","0","0","0"]
                obj = object_node(root,attr,[n_classes[int(classes)],x,y,int(x+w),int(y+h)],None)
                xml.appendChild(obj)

        else:

            for ann in data:
                classes = int(ann[0])
                seg = [round(float(val)*width,2) if idx%2 == 0 else round(float(val)*height,2) for idx,val in enumerate(ann[1:])]
                seg_x = seg[::2]
                seg_y = seg[1::2]
                x1, y1, x2, y2 = int(min(seg_x)), int(min(seg_y)), int(max(seg_x)), int(max(seg_y))
                attr = ["Unspecified","0","0","0"]
                obj = object_node(root,attr,[n_classes[classes],x1,y1,x2,y2],[seg_x,seg_y])
                xml.appendChild(obj)

        xml_str = root.toprettyxml(indent ="\t") 

        with open(os.path.join(destination,file[:-4]+".xml"),"w+") as f:
            f.write(xml_str[23:])
        
    print("\nThe conversion is completed.")

    print("\nValidation of converted annotation in progress..........")

    voc_validator(image_path, destination)

    print("\nThe annotated file is saved at '{}'.\n".format(destination))


from pycocotools.coco import COCO
import os
import tqdm
from xml.dom import minidom

from helpers_ai.utils.voc import element_node, child_node, object_node
from helpers_ai.utils.util import coco_rle2mask
from helpers_ai.utils.validator import coco_validator,voc_validator




def convert(json_path:str, image_path:str, destination:str) -> None:

    assert os.path.isdir(image_path) and os.path.isdir(destination), "image_path and destination must be a directory."
    assert os.path.isfile(json_path) and json_path.endswith(".json"), "json_path must be a path to a json file."

    coco = COCO(json_path)

    print("\nAnnotation Validator in progress.....")

    coco_validator(coco,image_path)

    imgid = coco.getImgIds()
    cat = coco.loadCats(coco.getCatIds())
    categ = [ct["name"] for ct in cat]

    print("")

    for imid in tqdm.tqdm(imgid):
        image = coco.loadImgs(imid)
        annid = coco.getAnnIds(imgIds=imid)
        annot = coco.loadAnns(annid)
        height, width = image[0]["height"],image[0]["width"]

        root = minidom.Document()
        xml = root.createElement("annotation")
        root.appendChild(xml)

        child_node(root,"folder","",xml)
        child_node(root,"filename",image[0]["file_name"],xml)
        child_node(root,"path",image[0]["file_name"],xml)

        source = element_node(root,"source",None)
        xml.appendChild(source)
        child_node(root,"database","",source)

        size = element_node(root,"size",None)
        xml.appendChild(size)
        child_node(root,"width",str(width),size)
        child_node(root,"height",str(height),size)
        child_node(root,"depth","3",size)

        child_node(root,"segmented","0",xml)

        for ann in annot:
            classes = ann["category_id"]
            ann_seg = ann["segmentation"]
            x, y, w, h = [int(i) for i in ann["bbox"]]
            attr = ["Unspecified","0","0","0"]

            if ann_seg == [[]]:  
                obj = object_node(root,attr,[categ[int(classes)],x,y,int(x+w),int(y+h)],None)
                xml.appendChild(obj)
         
            elif isinstance(ann_seg,dict):
                seg = coco_rle2mask(ann_seg["counts"], ann_seg["size"], 255, False, 50)
                
                for i in seg:
                    i = i.ravel()
                    obj = object_node(root,attr,[categ[int(classes)],x,y,int(x+w),int(y+h)],[i[::2], i[1::2]])
                    xml.appendChild(obj)
           
            else:
                seg_x = ann_seg[0][::2]
                seg_y = ann_seg[0][1::2]
                obj = object_node(root,attr,[categ[int(classes)],x,y,int(x+w),int(y+h)],[seg_x,seg_y])
                xml.appendChild(obj)
        
        xml_str = root.toprettyxml(indent ="\t") 

        with open(os.path.join(destination,image[0]["file_name"][:-4]+".xml"),"w+") as f:
            f.write(xml_str[23:])
        
    print("\nThe conversion is completed.")
    
    print("\nValidation of converted annotation in progress..........")

    voc_validator(image_path, destination)

    print("\nThe annotated file is saved at '{}'.\n".format(destination))


import tqdm
import cv2
import os
import json
from datetime import datetime
from pycocotools.coco import COCO

from helpers_ai.utils.validator import yolo_validator, coco_validator




def convert(labels_path:str, image_path:str, destination:str, n_classes:list[str]) -> None:

    print("\nAnnotation Validator in progress.....")

    yolo_validator(image_path, labels_path, n_classes)
    
    info = {"description": "YOLO 2 COCO Converted Dataset",
            "Contributer": "code63ReaPer",
            "date_created": datetime.strftime(datetime.today(),"%Y/%m/%d")}
    
    licenses = {}
    categories = []
    
    for idx,cls in enumerate(n_classes):
        categories.append({"id":idx,
                           "name":cls})
        
    images = []
    annotations = []
    ann_id = 0
    tot_dt = 0
    cl = []
    corrupted =[]

    for file in os.listdir(labels_path):
        
        with open(os.path.join(labels_path,file),"r") as lr:
            raw = lr.read()
            data = [sep.split(" ") for sep in raw.split("\n")]

        for ann in data:
            cl.append(ann[0])

            if (len(ann)==5) and (float(ann[1])>1 or float(ann[2])>1 or float(ann[3])>1 or float(ann[4])>1):
                corrupted.append(file)
                break

            elif len(ann)>5 and max([float(i) for i in ann[1:]])>1:
                corrupted.append(file)
                break

    if len(corrupted)>0:
        print("\nThere are {} corrupted files.\n".format(len(corrupted)))
    
        for i in corrupted:
            print(i,"data is corrupted.")
        
        print("")
        raise ValueError("Corrupted files are present")
    
    if len(set(cl)) != len(n_classes):
        raise AssertionError("The number of classes given does not match the classes in annotated file")
    
    print("")
    
    pbar = tqdm.tqdm(total=len(os.listdir(image_path)))
    
    for img_idx,file in enumerate(os.listdir(image_path)):
        img = cv2.imread(os.path.join(image_path,file))
        height,width = img.shape[:2]
        
        with open(os.path.join(labels_path,file[:-4]+".txt"),"r") as f:
            raw = f.read()
            data = [sep.split(" ") for sep in raw.split("\n")]
        
        images.append({"file_name": file,
                       "width": width,
                       "height": height,
                       "id": img_idx})
        tot_dt += len(data)
        
        if len(data[0])==5:
            
            for ann in data:
                classes, cx, cy, nw, nh = ann
                w, h = float(nw)*width, float(nh)*height
                x, y = round(((float(cx)*2*width)-w)/2,2), round(((float(cy)*2*height)-h)/2,2)
                annotations.append({"segmentation":[[]],
                                    "area": w*h,
                                    "iscrowd": 0,
                                    "image_id": img_idx,
                                    "bbox": [x, y, w, h],
                                    "category_id": int(classes),
                                    "id": ann_id})
                ann_id += 1
        
        else:

            for ann in data:
                classes = ann[0]
                seg = [round(float(val)*width,2) if idx%2 == 0 else round(float(val)*height,2) for idx,val in enumerate(ann[1:])]
                seg_x = seg[::2]
                seg_y = seg[1::2]
                x, y, w, h = round(min(seg_x),2), round(min(seg_y),2), round(max(seg_x)-min(seg_x),2), round(max(seg_y)-min(seg_y),2)
                annotations.append({"segmentation":[seg],
                                    "area": w*h,
                                    "iscrowd": 0,
                                    "image_id": img_idx,
                                    "bbox": [x, y, w, h],
                                    "category_id": int(classes),
                                    "id": ann_id})
                ann_id += 1
        
        pbar.update(img_idx+1)
    
    ann_data = {"info":info,
                "licenses":licenses,
                "images":images,
                "annotations":annotations,
                "categories":categories}
    
    print("\n\nSaving the json file...........\n")
    
    with open(os.path.join(destination,"yolo2coco.json"),"w+") as f1:
        json.dump(ann_data,f1)
    
    print("Save Completed.")

    print("\nThe conversion is completed.")

    print("\nValidation of converted annotation in progress..........\n")
    
    coco = COCO(os.path.join(destination,"yolo2coco.json"))
    catid = coco.getCatIds()
    annid = coco.getAnnIds()
    imgid = coco.getImgIds()

    coco_validator(coco, image_path)
    
    print("\nThere are {} images, {} annotations and {} classes are loaded.".format(img_idx+1,tot_dt,len(n_classes)))

    print("\nIn the saved json file there are {} images, {} annotations and {} categories are present.".format(len(imgid),len(annid),len(catid)))

    print("\nThe annotation file is saved at '{}' as 'yolo2coco.json'.\n".format(destination))


import json
import numpy as np
from sklearn.model_selection import train_test_split




def file_read(json_path:str, train_size:float, test_size:float, shuffle:bool) -> tuple[list,list]:

    assert train_size is not None or test_size is not None, "Either train_size or test_size may be None but not both."

    if train_size is not None and test_size is None:   
        assert isinstance(train_size, float) and train_size <= 1, "train_size must be float within [0-1]."
        test_size = round(abs(1-train_size), 2)

    elif test_size is not None and train_size is None:  
        assert isinstance(train_size, float) and train_size < 1, "test_size must be float and less than 1."
        train_size = round(abs(1-test_size), 2)

    assert (isinstance(train_size, float) and isinstance(test_size, float)) and ((train_size <= 1) and (test_size < 1)), "train_size must be within [0-1] and test_size less than 1 and both are must be in float."
    
    if (train_size+test_size) != 1:
        test_size = round(abs(1-train_size), 2) 

    with open(json_path, "r") as f:
        dt = json.load(f)

    train, test = train_test_split(dt, train_size=train_size, test_size=test_size, shuffle=shuffle)

    return (train, test)




def extractor(v, resize:bool, s_width:int, s_height:int) -> dict:
    img_name=v["ocr"].split("/")[-1]
    ann = v["poly"]
    text = v["transcription"]
    
    if isinstance(text,str):
        text = [v["transcription"]]
    
    w_scale = 1
    h_scale = 1
    p_lst = []
    b_lst = []

    for i in ann:
        width = i["original_width"]
        height = i["original_height"]

        if resize:
            w_scale = s_width/width
            h_scale = s_height/height
            width = s_width
            height = s_height
        
        poly = np.array(i["points"])
        seg_x =(poly[:,0] * w_scale) / 100 * width
        seg_y =(poly[:,1] * h_scale) / 100 * height

        xmin= int(np.min(seg_x))
        ymin= int(np.min(seg_y))
        xmax= int(np.max(seg_x))
        ymax= int(np.max(seg_y))
        
        poly = np.array([[x,y] for x,y in zip(seg_x,seg_y)]).ravel()
        poly = [int(i) for i in poly]

        p_lst.append(poly)
        b_lst.append([xmin,ymin,xmax,ymax])
    
    return {"p_lst": p_lst,
            "b_lst": b_lst,
            "t_lst": text,
            "image_name": img_name,
            "width": width,
            "height": height,
            "len":len(ann)}


from pycocotools.coco import COCO
import os
import tqdm

from helpers_ai.utils.util import coco_rle2mask
from helpers_ai.utils.validator import coco_validator, yolo_validator




def __check_corrupt(corrupted):

    if len(corrupted)>0:

        print("\nThere are {} corrupted files.\n".format(len(corrupted)))
    
        for i in corrupted:
            print(i,"data is corrupted.")
            
        print("")
        
        raise ValueError("Corrupted files are present")




def detetcion(json_path:str, image_path:str, destination:str) -> None:
    
    coco = COCO(json_path)

    print("\nAnnotation Validator in progress.....")

    coco_validator(coco,image_path)

    imgid = coco.getImgIds()
    cat = coco.loadCats(coco.getCatIds())
    categ = [ct["name"] for ct in cat]
    corrupted = []

    print("")

    for imid in tqdm.tqdm(imgid):
        image = coco.loadImgs(imid)
        annid = coco.getAnnIds(imgIds=imid)
        annot = coco.loadAnns(annid)
        height, width = image[0]["height"],image[0]["width"]
        out = []
        ps = True
        
        for ann in annot:
            classes = ann["category_id"]
            x, y, w, h = [float(i) for i in ann["bbox"]]
            cx, cy, nw, nh = (x+x+w)/(2*width), (y+y+h)/(2*height), w/width, h/height
            
            if cx>1 or cy>1 or nw>1 or nh>1 :
                corrupted.append(image[0]["file_name"][:-4]+".txt")
                ps = False
                break
            
            out.append(f"{classes} {cx} {cy} {nw} {nh}")

        if ps:
            with open(os.path.join(destination,image[0]["file_name"][:-4]+".txt"),"w+") as f:
                f.writelines("\n".join(out)) 

    __check_corrupt(corrupted)

    print("\nThe conversion is completed.")

    print("\nValidation of converted annotation in progress..........\n")

    yolo_validator(image_path,destination,categ)

    with open(os.path.join(destination,"__classes.txt"),"w+") as f:
            f.writelines("\n".join(categ))

    print("\nThe annotation file is saved at '{}'.\n".format(destination))
   
                


def segmentation(json_path:str, image_path:str, destination:str) -> None:
    
    coco = COCO(json_path)

    print("\nAnnotation Validator in progress.....")

    coco_validator(coco,image_path)
    
    imgid = coco.getImgIds()
    cat = coco.loadCats(coco.getCatIds())
    categ = [ct["name"] for ct in cat]
    corrupted = []

    print("")

    for imid in tqdm.tqdm(imgid):
        image = coco.loadImgs(imid)
        annid = coco.getAnnIds(imgIds=imid)
        annot = coco.loadAnns(annid)
        height, width = image[0]["height"],image[0]["width"]
        out = []
        ps = True

        
        for ann in annot:
            classes = ann["category_id"]
            ann_seg = ann["segmentation"]

            if isinstance(ann_seg,dict):
                seg = coco_rle2mask(ann_seg["counts"], ann_seg["size"], 255, False, 50)

                for i in seg:
                    seg_p = [val/width if idx%2 == 0 else val/height for idx,val in enumerate(i.ravel())]

                    if max(seg_p)>1:
                        corrupted.append(image[0]["file_name"][:-4]+".txt")
                        ps = False
                        break

                    out.append(f"{classes} {' '.join([str(i) for i in seg_p])}")

            else:
                seg = [val/width if idx%2 == 0 else val/height for idx,val in enumerate(ann_seg[0])]

                if max(seg)>1:
                    corrupted.append(image[0]["file_name"][:-4]+".txt")
                    ps = False
                    break
                
                out.append(f"{classes} {' '.join([str(i) for i in seg])}")
            
        if ps:
            with open(os.path.join(destination,image[0]["file_name"][:-4]+".txt"),"w+") as f:
                f.writelines("\n".join(out))
    
    __check_corrupt(corrupted)

    print("\nThe conversion is completed.")

    print("\nValidation of converted annotation in progress..........\n")

    yolo_validator(image_path,destination,categ)

    with open(os.path.join(destination,"__classes.txt"),"w+") as f:
            f.writelines("\n".join(categ))

    print("\nThe conversion is completed.")
    print("\nThe annotation file is saved at '{}'.\n".format(destination))


import os
import cv2
import yaml
import tqdm
import imageio as iio
from pycocotools.coco import COCO
import xml.etree.ElementTree as et

from helpers_ai.utils.validator import coco_validator, yolo_validator, voc_validator
from helpers_ai.utils.util import annot_bbox, annot_seg, coco_rle2mask, annot_color




def coco_annotator(image_folder:str, json_path:str, destination_path:str, segmentation:bool,
                   bbox:bool, class_names:tuple[str], file_name:str, size:tuple[int,int],
                   video_out:bool, fps:int, gif_out:bool, duration_of_frame:int, loop:int) -> None:
    
    assert (bbox and segmentation) or bbox or segmentation, "Either segmentation or bbox must be True or both may be True."
    assert json_path.endswith(".json") and os.path.isfile(json_path), "annotation file must be a json file."
    assert os.path.isdir(image_folder), "image_folder must be a folder."
    assert file_name[-4:] in [".mp4",".gif"], "video_name must be in .gif or .mp4 format."

    coco = COCO(json_path)

    print("\nAnnotation Validator in progress.....")

    coco_validator(coco,image_folder)

    print("\nAnnotation in progress...........")
    
    if class_names != ():
        imgid = coco.getImgIds()
        cat_id = coco.getCatIds()
    else:
        cat_id = coco.getCatIds(catNms=class_names)
        imgid = coco.getImgIds(catIds=cat_id)

    if video_out:
        vid_wr = cv2.VideoWriter(os.path.join(destination_path, file_name[:-4]+".mp4"), cv2.VideoWriter.fourcc(*"mp4v"),
                                 fps, size)
    
    elif gif_out:
        gif_wr = iio.get_writer(os.path.join(destination_path,file_name[:-4]+".gif"), mode="I", duration=duration_of_frame, loop=loop) 

    print("")

    for imid in tqdm.tqdm(imgid):
        image = coco.loadImgs(imid)
        annot = coco.loadAnns(coco.getAnnIds(imgIds=imid, catIds=cat_id))
        img = os.path.join(image_folder,image[0]["file_name"])

        for ann in annot:
            classes = int(ann["category_id"])
            ann_seg = ann["segmentation"]
            ann_bbox =  ann["bbox"]

            if segmentation and ann_seg != [[]]:
                if isinstance(ann_seg, dict):
                    continue
                    seg_lst = coco_rle2mask(ann_seg["counts"], ann_seg["size"], 255, False, 50)
                    for k in seg_lst:
                        seg = k.ravel()
                        print(seg)
                        seg_x = seg[0][::2]
                        seg_y = seg[0][1::2]
                        img = annot_seg(img, [[x,y] for x,y in zip(seg_x,seg_y)], annot_color(classes))
                else:
                    seg_x = ann_seg[0][::2]
                    seg_y = ann_seg[0][1::2]
                    img = annot_seg(img, [[x,y] for x,y in zip(seg_x,seg_y)], annot_color(classes))

            if bbox and ann_bbox != []:
                x,y,w,h = [int(i) for i in ann_bbox]
                img = annot_bbox(img, [x, y, x+w, y+h], annot_color(classes), 2, coco.loadCats(classes)[0]["name"], 1)
        
        if video_out:
            img = cv2.resize(img, size)
            vid_wr.write(img)
        
        elif gif_out:
            img = cv2.resize(img, size)
            gif_wr.append_data(cv2.cvtColor(img,cv2.COLOR_BGR2RGB))
        
        else:
            cv2.imwrite(os.path.join(destination_path, image[0]["file_name"]), img)
    
    if video_out:
        vid_wr.release()
    elif gif_out:
        gif_wr.close()

    print("\nSave Completed.")

    print("\nThe annotation process is Completed.")

    if video_out:
        print("\nThe files are saved at '{}' named '{}'.".format(destination_path, file_name[:-4]+".mp4"))
    elif gif_out:
        print("\nThe files are saved at '{}' named '{}'.".format(destination_path, file_name[:-4]+".gif"))
    else:
        print("\nThe files are saved at '{}'.".format(destination_path))



def yolo_annotator(image_folder:str, labels_folder:str, cls_file_path:str, destination_path:str, segmentation:bool,
                   bbox:bool, class_names:tuple[str], file_name:str, size:tuple[int,int], video_out:bool, fps:int, 
                   gif_out:bool, duration_of_frame:int, loop:int) -> None:
    
    assert (bbox and segmentation) or bbox or segmentation, "Either segmentation or bbox must be True or both may be True."
    assert os.path.isfile(cls_file_path), "cls_file_path must be a class file in txt or yaml format."
    assert os.path.isdir(image_folder) and os.path.isdir(labels_folder), "images_folder and labels_folder must be a directory."
    assert file_name[-4:] in [".mp4",".gif"], "video_name must be in .gif or .mp4 format."

    print("\nAnnotation Validator in progress.....")

    with open(cls_file_path, "r") as f:
        if cls_file_path.endswith(".yaml"):
            cls_name = yaml.safe_load(f)["names"]
        else:
            cls_name = f.readlines()

    yolo_validator(image_folder, labels_folder, list(range(len(cls_name))))

    print("\nAnnotation in progress...........\n")

    
    
    if video_out:
        vid_wr = cv2.VideoWriter(os.path.join(destination_path, file_name[:-4]+".mp4"), cv2.VideoWriter.fourcc(*"mp4v"),
                                 fps, size)
    
    elif gif_out:
        gif_wr = iio.get_writer(os.path.join(destination_path,file_name[:-4]+".gif"), mode="I", duration=duration_of_frame, loop=loop) 

       
    for file in tqdm.tqdm(os.listdir(image_folder)):

        img = cv2.imread(os.path.join(image_folder, file))
        height, width = img.shape[:2]

        with open(os.path.join(labels_folder, file[:-4]+".txt"), "r") as f:
            data = [i.split() for i in f.readlines()]

        for i in data:
            cls = int(i[0])
            
            if class_names != () and cls_name[cls] not in class_names:  
                continue

            ann_data = i[1:]
            seg = [round(float(val)*width,2) if idx%2 == 0 else round(float(val)*height,2) for idx,val in enumerate(ann_data)]
            seg_x = seg[::2]
            seg_y = seg[1::2]

            if segmentation and len(ann_data) > 4:
                img = annot_seg(img, [[x,y] for x,y in zip(seg_x,seg_y)], annot_color(cls))
           
            if bbox and len(ann_data) > 3:
                x, y, x1, y1 = [int(min(seg_x)), int(min(seg_y)), int(max(seg_x)), int(max(seg_y))]
                img = annot_bbox(img, [x, y, x1, y1], annot_color(cls), 2, cls_name[cls], 1)

        if video_out:
            img = cv2.resize(img, size)
            vid_wr.write(img)

        elif gif_out:
            img = cv2.resize(img, size)
            gif_wr.append_data(cv2.cvtColor(img,cv2.COLOR_BGR2RGB))

        else:
            cv2.imwrite(os.path.join(destination_path, file), img)
    
    if video_out:
        vid_wr.release()
    elif gif_out:
        gif_wr.close()

    print("\nSave Completed.")

    print("\nThe annotation process is Completed.")

    if video_out:
        print("\nThe files are saved at '{}' named '{}'.".format(destination_path, file_name[:-4]+".mp4"))
    elif gif_out:
        print("\nThe files are saved at '{}' named '{}'.".format(destination_path, file_name[:-4]+".gif"))
    else:
        print("\nThe files are saved at '{}'.".format(destination_path))
            
        


def voc_annotator(image_folder:str, labels_folder:str, destination_path:str, segmentation:bool,
                  bbox:bool, class_names:tuple[str], file_name:str, size:tuple[int,int], video_out:bool, fps:int,
                  gif_out:bool, duration_of_frame:int, loop:int) -> None:
        
    assert (bbox and segmentation) or bbox or segmentation, "Either segmentation or bbox must be True or both may be True."
    assert os.path.isdir(image_folder) and os.path.isdir(labels_folder), "images_folder and labels_folder must be a directory."
    assert file_name[-4:] in [".mp4",".gif"], "video_name must be in .gif or .mp4 format."

    print("\nAnnotation Validator in progress.....")

    cls = voc_validator(image_folder,labels_folder)

    print("\nAnnotation in progress...........\n")

    if video_out:
        vid_wr = cv2.VideoWriter(os.path.join(destination_path, file_name[:-4]+".mp4"), cv2.VideoWriter.fourcc(*"mp4v"),
                                 fps, size)
    
    elif gif_out:
        gif_wr = iio.get_writer(os.path.join(destination_path,file_name[:-4]+".gif"), mode="I", duration=duration_of_frame, loop=loop) 


    for file in tqdm.tqdm(os.listdir(labels_folder)):
        if file.endswith(".xml"):
            tree = et.parse(os.path.join(labels_folder,file))
            f_name = tree.find("filename").text
            img = cv2.imread(os.path.join(image_folder, f_name))
                
            for ann in tree.findall("object"):
                classes = ann.find("name").text

                if class_names != () and classes not in class_names:
                    continue

                if segmentation and ann.find("polygon") is not None:
                    seg = [float(child.text) for child in ann.find("polygon")]
                    seg_x = seg[::2]
                    seg_y = seg[1::2]
                    img = annot_seg(img, [[x,y] for x,y in zip(seg_x,seg_y)], annot_color(cls.index(classes)))

                rd_bbox = ann.find("bndbox")
                if bbox and rd_bbox is not None:
                    xmin = int(rd_bbox.find("xmin").text)
                    ymin = int(rd_bbox.find("ymin").text)
                    xmax = int(rd_bbox.find("xmax").text)
                    ymax = int(rd_bbox.find("ymax").text)
                    img = annot_bbox(img, [xmin, ymin, xmax, ymax], annot_color(cls.index(classes)), 2, classes, 1)

            if video_out:
                img = cv2.resize(img, size)
                vid_wr.write(img)
            
            elif gif_out:
                img = cv2.resize(img, size)
                gif_wr.append_data(cv2.cvtColor(img,cv2.COLOR_BGR2RGB))

            else:
                cv2.imwrite(os.path.join(destination_path, f_name), img)
    
    if video_out:
        vid_wr.release()
    elif gif_out:
        gif_wr.close()
    
    print("\nSave Completed.")

    print("\nThe annotation process is Completed.")

    if video_out:
        print("\nThe files are saved at '{}' named '{}'.".format(destination_path, file_name[:-4]+".mp4"))
    elif gif_out:
        print("\nThe files are saved at '{}' named '{}'.".format(destination_path, file_name[:-4]+".gif"))
    else:
        print("\nThe files are saved at '{}'.".format(destination_path))


import argparse
import yaml

from helpers_ai import __version__
from helpers_ai.annotator import annotate, image_convertor, video_convertor
from helpers_ai.dataset_convertors import coco2voc, coco2yolo, yolo2coco, yolo2voc, voc2coco, voc2yolo, jm2mmocr


def cli(): 
    """Main entry point for the Helper."""
    parser = argparse.ArgumentParser(description="Helpers tool which helps in the progress of making models.")
    parser.add_argument('--version', action='version', version=f'Helpers-ai v{__version__}')
    subparsers = parser.add_subparsers(title='subcommands', dest='subcommand', description="Available subcommands in Helpers-ai:")

    annot_parser = subparsers.add_parser(
        'annot',
        help='Manage the Annotation Process',
        description='Annotate images with the provided data. The supported annotation formats are "COCO", "YOLO" and "Pascal_VOC" formats.'
    )

    annot_parser.add_argument("format", type=str, choices=['YOLO', 'COCO', 'VOC'], metavar= "Dataset_format",
                            help="Specify the dataset format you want to annotate. Available formats 'YOLO', 'COCO', 'VOC'")
    annot_parser.add_argument("image_path", type=str, metavar= "Image_path", help = 'Local image path directory')
    annot_parser.add_argument("data", nargs="*", metavar= "labels_path", help = 'Local label file path')
    annot_parser.add_argument("-des", "--dest", type=str, metavar= "", help = 'Desination path', default=".")
    annot_parser.add_argument("-bb", "--bbox", type=str, metavar="", help="Perform box annotation process", default=True)
    annot_parser.add_argument("-cls", nargs="*", metavar="", help="Perform annotation for specified classes", default=())
    annot_parser.add_argument("-o", "--out_name", type=str, metavar="", help="Output filename", default="output.mp4")
    annot_parser.add_argument("-lp", "--loop", type=int, metavar="", help="Times of loop for gif image. 0 for infinie loop", default=0)
    annot_parser.add_argument("-vid", "--video", action="store_true", help="Converts the annotated images into video")
    annot_parser.add_argument("--gif", action="store_true", help="Converts the annotated images into a gif file")
    annot_parser.add_argument("-seg", "--segment", action="store_true", help="Perform Segment annotation process")
    annot_parser.add_argument("--resize", type=int, nargs=2, metavar="", help="Resize the image to the specified size", default=(640,640))
    annot_parser.add_argument("--fps", type=int, metavar="", help="Output video fps", default=24)
    annot_parser.add_argument("--frame-dur", type=int, metavar="", help="Frame duration for gif image", default=700)



    img_conv_parser = subparsers.add_parser(
        'image-convert',
        help='Manage the Image conversion Process',
        description='Convert the images into video or gif image'
    )

    img_conv_parser.add_argument("task", type=str, choices=['img2vid', 'img2gif'], metavar="Task",
                                help="Specify the task you want to perform. Available task are 'img2vid', 'img2gif'")
    img_conv_parser.add_argument("image_path", type=str, metavar= "Image_path", help = 'Local image path directory')
    img_conv_parser.add_argument("-des", "--dest", type=str, metavar= "", help = 'Desination path', default=".")
    img_conv_parser.add_argument("-o", "--out_name", type=str, metavar="", help="Output filename", default="output.mp4")
    img_conv_parser.add_argument("--resize",type=int, nargs=2, metavar="", help="Resize the image to the specified size", default=(640,640))
    img_conv_parser.add_argument("--fps", type=int, metavar="", help="Output video fps", default=24)
    img_conv_parser.add_argument("--frame-dur", type=int, metavar="", help="Frame duration for gif image", default=700)
    img_conv_parser.add_argument("-lp", "--loop", type=int, metavar="", help="Times of loop for gif image. 0 for infinie loop", default=0)



    vid_conv_parser = subparsers.add_parser(
        'video-convert',
        help='Manage the video conversion Process',
        description='Convert the video into images, gif or change the fps of the video or crop the video.'
    )

    vid_conv_parser.add_argument("task", type=str, choices= ['vid2img', 'vid2gif', 'change_fps', 'cropper'], metavar="Task",
                    help="Specify the task you want to perform. Available task are 'vid2img', 'vid2gif', 'change_fps', 'cropper'")
    vid_conv_parser.add_argument("video_path", type=str, metavar= "video_path", help = 'Local video file path')
    vid_conv_parser.add_argument("-des", "--dest", type=str, metavar= "", help = 'Desination path', default=".")
    vid_conv_parser.add_argument("-o", "--out_name", type=str, metavar="", help="Output filename", default=None)
    vid_conv_parser.add_argument("--resize", type=int, nargs=2, metavar="", help="Resize the image to the specified size", default=(640,640))
    vid_conv_parser.add_argument("--frame-dur", type=int, metavar="", help="Frame duration for gif image", default=700)
    vid_conv_parser.add_argument("-lp", "--loop", type=int, metavar="", help="Times of loop for gif image. 0 for infinie loop", default=0)
    vid_conv_parser.add_argument("--fps", type=int, metavar="", help="Output video fps", default=24)
    vid_conv_parser.add_argument("--fast", action="store_true", help="Increase the speed of the video playback")
    vid_conv_parser.add_argument("-st", "--start-time", metavar="", help="Start-time of the video to crop. It must be in 'h:m:s' format", default="0:0:0")
    vid_conv_parser.add_argument("-et", "--end-time", metavar="", help="End-time of the video to crop. It must be in 'h:m:s' format", default="0:0:1")



    dat_conv_parser = subparsers.add_parser(
        'dataset-convert',
        help='Manage the Dataset conversion Process',
        description='Convert the dataset from one format to other format like COCO, YOLO, VOC.'
    )

    dat_conv_parser.add_argument("task", type=str, metavar="Task",
                                choices= ['coco2voc', 'coco2yolo', 'voc2coco', 'voc2yolo', 'yolo2coco', 'yolo2voc', 'jm2mmocr'], 
                                help="Specify the task you want to perform. Available task are 'coco2voc', 'coco2yolo', 'voc2coco', 'voc2yolo', 'yolo2coco', 'yolo2voc', 'jm2mmocr'")
    dat_conv_parser.add_argument("image_path", type=str, metavar="Image_path", help ='Local image path directory')
    dat_conv_parser.add_argument("data", nargs="*", metavar= "labels_path", help = 'Local label file path')
    dat_conv_parser.add_argument("-des", "--dest", type=str, metavar= "", help = 'Desination path', default=".")
    dat_conv_parser.add_argument("-m", "--mode", type=str, choices=['detect', 'segment'], metavar="",
                                help="Selection of mode either detect or segment", default="detect")




    args = parser.parse_args()



    if args.subcommand == "annot":

        assert len(args.data) < 3, "Labels_path accepts a max of two arguments only."
        
        if args.format.lower() == "coco":
            annotate.coco_annotator(args.image_path, args.data[0], args.dest, args.segment, eval(args.bbox), args.cls, args.out_name,
                                    args.resize, args.video, args.fps, args.gif, args.frame_dur, args.loop)
        
        elif args.format.lower() == "yolo":
            annotate.yolo_annotator(args.image_path, args.data[0], args.data[1], args.dest, args.segment, eval(args.bbox), args.cls, args.out_name,
                                    args.resize, args.video, args.fps, args.gif, args.frame_dur, args.loop)
        
        elif args.format.lower() == "voc":
            annotate.voc_annotator(args.image_path, args.data[0], args.dest, args.segment, eval(args.bbox), args.cls, args.out_name,
                                args.resize, args.video, args.fps, args.gif, args.frame_dur, args.loop)



    elif args.subcommand == "image-convert":
        if args.task.lower() == "img2vid":
            image_convertor.img2vid(args.image_path, args.dest, args.out_name, args.fps, args.resize)
        
        elif args.task.lower() == "img2gif":
            image_convertor.img2gif(args.image_path, args.dest, args.out_name, args.resize, args.frame_dur, args.loop)



    elif args.subcommand == "video-convert":
        print(args.task.lower())
        if args.task.lower() == "vid2img":
            video_convertor.vid2img(args.video_path, args.dest, args.resize)
        
        elif args.task.lower() == "vid2gif":
            video_convertor.vid2gif(args.video_path, args.dest, args.out_name, args.resize, args.frame_dur, args.loop)
        
        elif args.task.lower() == "change_fps":
            video_convertor.fps_changer(args.video_path, args.fps, args.dest, args.out_name, args.resize, args.fast)

        elif args.task.lower() == "cropper":
            video_convertor.vid_cropper(args.video_path, args.start_time, args.end_time, args.dest, args.out_name, args.fps, args.resize)



    elif args.subcommand == "dataset-convert":
        if args.task.lower() == "coco2voc":
            coco2voc.convert(args.data[0], args.image_path, args.dest)
        
        elif args.task.lower() == "coco2yolo":
            if args.mode == "detect":
                coco2yolo.detetcion(args.data[0], args.image_path, args.dest)
            elif args.mode == "segment":
                coco2yolo.segmentation(args.data[0], args.image_path, args.dest)
        
        elif args.task.lower() == "yolo2coco":
            
            assert len(args.data) == 2, "Provide label directory and class file path in txt or yaml format."

            with open(args.data[1], "r") as f:
                if args.data[1].endswith(".yaml"):
                    cls_name = yaml.safe_load(f)["names"]
                else:
                    cls_name = f.readlines()

            yolo2coco.convert(args.data[0], args.image_path, args.dest, cls_name)
        
        elif args.task.lower() == "yolo2voc":

            assert len(args.data) == 2, "Provide label directory and class file path in txt or yaml format."

            with open(args.data[1], "r") as f:
                if args.data[1].endswith(".yaml"):
                    cls_name = yaml.safe_load(f)["names"]
                else:
                    cls_name = f.readlines()
            
            yolo2voc.convert(args.data[0], args.image_path, args.dest, cls_name)

        elif args.task.lower() == "voc2coco":
            voc2coco.convert(args.data[0], args.image_path, args.dest)
        
        elif args.task.lower() == "voc2yolo":
            if args.mode == "detect":
                voc2yolo.detection(args.data[0], args.image_path, args.dest)
            elif args.mode == "segment":
                voc2yolo.segment(args.data[0], args.image_path, args.dest)


if __name__ == "__main__":
    cli()


import json 
import os
import tqdm
import cv2
import shutil

from helpers_ai.utils.jm2mm import extractor, file_read




def detect(json_path:str, image_path:str, destination_path:str, train_size:float=None,
            test_size:float=0.1, resize:bool=False, size:tuple[int,int]=(640,640), shuffle:bool=True) -> None:

    assert len(size) == 2, "size must have a length of 2."
    
    meta_data = {"dataset_type": "OCRDataset",
                "task_name": "textdet",
                "category":{"id":0, "name":"text"}
                }
    
    spt = file_read(json_path, train_size, test_size, shuffle)

    try:
        os.mkdir(os.path.join(destination_path, "detect"))
    except(FileExistsError):
        print("\nFolder detect is already exists.")

    try:
        os.mkdir(os.path.join(destination_path,"detect","annotations"))
    except(FileExistsError):
        print("\nFolder annotations is already exists.")

    for idx, phase in enumerate(spt):
        ann_con = []
        if idx == 0:
            ph = "train"
        elif idx == 1:
            ph = "val"
        try:
            print("\nCreating {} annotations..........".format(ph))
            os.mkdir(os.path.join(destination_path,"detect",ph))
        except(FileExistsError):
            print("\nFolder {} is already exists.".format(ph))
        
        print("")

        for v in tqdm.tqdm(phase):
            b_data = []
            img_name=v["ocr"].split("/")[-1]

            if resize:
                img = cv2.imread(os.path.join(image_path,img_name))
                img = cv2.resize(img, size)
                cv2.imwrite(os.path.join(destination_path,"detect",ph,img_name), img)
            else:
                shutil.copy(os.path.join(image_path,img_name),os.path.join(destination_path,"detect",ph,img_name))

            dt = extractor(v, resize, size[0], size[1])
            
            for poly, bbox, text in zip(dt["p_lst"], dt["b_lst"], dt["t_lst"]):

                b_data.append({"polygon": poly,
                            "bbox": bbox,
                            "bbox_label": 0,
                            "text": text,
                            "ignore": False
                            })
            ann_con.append({"img_path": dt["image_name"],
                        "height": dt["height"],
                        "width": dt["width"],
                        "instances": b_data})
            
        annot = {"metainfo":meta_data,
                    "data_list": ann_con}
        print("\n{} Dataset contains".format(ph), len(annot["data_list"]), "image annotations.")

        with open(os.path.join(destination_path,"detect","annotations","mmocr_detect_"+ph+".json"),"w") as f:
                json.dump(annot, f, indent=2)

    print("\nConversion Successful.\n")




def recog(json_path:str, image_path:str, destination_path:str, train_size:float=None,
            test_size:float=0.1, resize:bool=False, size:tuple[int,int]=(640,640), shuffle:bool=True) -> None:

    assert len(size) == 2, "size must have a length of 2."

    spt = file_read(json_path, train_size, test_size, shuffle)

    try:
        os.mkdir(os.path.join(destination_path,"recog"))
    except(FileExistsError):
        print("\nFolder recog is already exists.")

    try:
        os.mkdir(os.path.join(destination_path,"recog","annotations"))
    except(FileExistsError):
        print("\nFolder annotations is already exists.")

    for idx, phase in enumerate(spt):
        ann_con = []

        if idx == 0:
            ph = "train"
        elif idx == 1:
            ph = "val"

        try:
            print("\nCreating {} annotations..........".format(ph))
            os.mkdir(os.path.join(destination_path,"recog",ph))
        except(FileExistsError):
            print("\nFolder {} is already exists.".format(ph))
        
        print("")
        count = 0
        for v in tqdm.tqdm(phase):

            dt = extractor(v, image_path, destination_path, ph, "recog", resize, size[0], size[1])

            count += dt["len"]
            img = cv2.imread(os.path.join(image_path,dt["image_name"]))
            if resize:
                img = cv2.resize(img, size)

            for idx1, (bbox, text) in enumerate(zip(dt["b_lst"], dt["t_lst"])):

                cr_img = img[bbox[1]:bbox[3],bbox[0]:bbox[2]]
                cv2.imwrite(os.path.join(destination_path,"recog",ph,dt["image_name"][:-4]+"-"+str(idx1+1)+".png"),cr_img)
                ann_con.append(f"{dt['image_name'][:-4]+'-'+str(idx1+1)+'.png'} {text}")
        print(count)        
        with open(os.path.join(destination_path,"recog","annotations","mmocr_recog_"+ph+".txt"),"w") as f:
                f.writelines("\n".join(ann_con))
    
    print("\nConversion Successful.\n")




def kie_closed(json_path:str, image_path:str, destination_path:str, train_size:float=None,
            test_size:float=0.1, resize:bool=False, size:tuple[int,int]=(640,640), shuffle:bool=True) -> None:

    assert len(size) == 2, "size must have a length of 2."

    spt = file_read(json_path, train_size, test_size, shuffle)

    try:
        os.mkdir(os.path.join(destination_path,"kie-closed"))
    except(FileExistsError):
        print("\nFolder kie-closed is already exists.")

    try:
        os.mkdir(os.path.join(destination_path,"kie-closed","annotations"))
    except(FileExistsError):
        print("\nFolder annotations is already exists.")

    label = []
    for phase in spt:
        for v in phase:
            ann = v["label"]
            for idx1,i in enumerate(ann):
                label.append(i["labels"][0])
    cat = sorted(list(set(label)))

    with open(os.path.join(destination_path,"kie-closed","annotations","mmocr_kie_metainfo.txt"),"w") as f:
        for idx2, i in enumerate(cat):
            f.writelines(f"{idx2} {i}\n")

    for idx, phase in enumerate(spt):
        ann_con = []
        label = []

        if idx == 0:
            ph = "train"
        elif idx == 1:
            ph = "val"

        try:
            print("\nCreating {} annotations..........".format(ph))
            os.mkdir(os.path.join(destination_path,"kie-closed",ph))
        except(FileExistsError):
            print("\nFolder {} is already exists.".format(ph))
        
        print("")

        for v in tqdm.tqdm(phase):

            b_data = []
            img_name=v["ocr"].split("/")[-1]

            if resize:
                img = cv2.imread(os.path.join(image_path,img_name))
                img = cv2.resize(img, size)
                cv2.imwrite(os.path.join(destination_path,"kie-closed",ph,img_name), img)
            else:
                shutil.copy(os.path.join(image_path,img_name),os.path.join(destination_path,"kie-closed",ph,img_name))

            dt = extractor(v, resize, size[0], size[1])

            for idx1, (bbox, text) in enumerate(zip(dt["b_lst"], dt["t_lst"])):

                b_data.append({
                            "box": [bbox[0],bbox[1],bbox[0],bbox[3],bbox[2],bbox[3],bbox[2],bbox[1]],
                            "text": text,
                            "label": cat.index(v["label"][idx1]["labels"][0])})
            ann_con.append({"file_name": dt["image_name"],
                        "height": dt["height"],
                        "width": dt["width"],
                        "annotations": b_data})
            
        print("\n{} Dataset contains".format(ph), len(ann_con), "image annotations.")

        with open(os.path.join(destination_path,"kie-closed","annotations","mmocr_kie_"+ph+".json"),"w") as f:
            json.dump(ann_con, f, indent=2)
          
        print("\nConversion Successful.\n")


import xml.etree.ElementTree as et
import tqdm
import os
import json
from datetime import datetime
from pycocotools.coco import COCO

from helpers_ai.utils.validator import voc_validator, coco_validator, class_getter



    
def convert(label_path:str, image_path:str, destination:str) -> None:

    print("\nAnnotation Validator in progress.....")

    voc_validator(image_path,label_path)

    info = {"description": "VOC 2 COCO Converted Dataset",
            "Contributer": "code63ReaPer",
            "date_created": datetime.strftime(datetime.today(),"%Y/%m/%d")}
    
    licenses = {}
    categories = []
    s_cls = class_getter(label_path)

    for idx,cls in enumerate(s_cls):
        categories.append({"id":idx,
                           "name":cls})
        
    images = []
    annotations = []
    ann_id = 0
    tot_dt = 0
    img_id = 0
    
    print("\nThere are {} classes are present in the given labels_path.".format(len(s_cls)))
    flg = input("\nContinue[y/n]: ")
    print("")

    if flg == "n" or flg == "N":
        return

    elif flg == "y" or flg =="Y":
        for file in tqdm.tqdm(os.listdir(label_path)):
            if file.endswith(".xml"):
                tree = et.parse(os.path.join(label_path,file))
                filename = tree.find("filename").text
                size = tree.find("size")
                width = int(size.find("width").text)
                height = int(size.find("height").text)

                images.append({"file_name": filename,
                       "width": width,
                       "height": height,
                       "id": img_id})
                
                tot_dt += len(tree.findall("object"))
                    
                for idx,ann in enumerate(tree.findall("object")):
                    classes = ann.find("name").text
                    bbox = ann.find("bndbox")
                    xmin = int(bbox.find("xmin").text)
                    ymin = int(bbox.find("ymin").text)
                    xmax = int(bbox.find("xmax").text)
                    ymax = int(bbox.find("ymax").text)
                    x, y, w, h = xmin, ymin, xmax-xmin, ymax-ymin
                    
                    if ann.find("polygon") is not None:      
                        seg = [float(child.text) for child in ann.find("polygon")]
                        annotations.append({"segmentation":[seg],
                                    "area": w*h,
                                    "iscrowd": 0,
                                    "image_id": img_id,
                                    "bbox": [x, y, w, h],
                                    "category_id": s_cls.index(classes),
                                    "id": ann_id})
                        ann_id += 1
                   
                    else:
                        annotations.append({"segmentation":[[]],
                                    "area": w*h,
                                    "iscrowd": 0,
                                    "image_id": img_id,
                                    "bbox": [x, y, w, h],
                                    "category_id": s_cls.index(classes),
                                    "id": ann_id})
                        ann_id += 1
                
                img_id += 1
            
    ann_data = {"info":info,
                "licenses":licenses,
                "images":images,
                "annotations":annotations,
                "categories":categories}
    
    print("\nSaving the json file...........\n")
    
    with open(os.path.join(destination,"voc2coco.json"),"w+") as f1:
        json.dump(ann_data,f1)
    
    print("Save Completed.")

    print("\nThe conversion is completed.")

    print("\nValidation of converted annotation in progress..........\n")
    
    coco = COCO(os.path.join(destination,"voc2coco.json"))
    catid = coco.getCatIds()
    annid = coco.getAnnIds()
    imgid = coco.getImgIds()

    coco_validator(coco, image_path)
    
    print("\nThere are {} images, {} annotations and {} classes are loaded.".format(img_id,tot_dt,len(s_cls)))

    print("\nIn the saved json file there are {} images, {} annotations and {} categories are present.".format(len(imgid),len(annid),len(catid)))

    print("\nThe annotation file is saved at '{}' as 'voc2coco.json'.\n".format(destination))


import os
import cv2
import tqdm
import natsort
import imageio as iio




def img2vid(images_path:str, destination_path:str, filename:str,
             fps:int, size:tuple[int,int]) -> None:

    assert os.path.isdir(images_path), "image_path must be a directory."
    assert len(size) == 2, "size must contain two integers."
    # assert len(filename.split(".")) == 1, "filename must not contain extensions."

    vid_wr = cv2.VideoWriter(os.path.join(destination_path,filename[:-4]+".mp4"), cv2.VideoWriter.fourcc(*"mp4v"), fps=fps, frameSize=size)

    for file in tqdm.tqdm(natsort.natsorted(os.listdir(images_path))):
        if file.endswith((".jpg",".png",".jpeg")):
            img = cv2.imread(os.path.join(images_path, file))
            img = cv2.resize(img, size)
            vid_wr.write(img)
    
    vid_wr.release()

    print("\nConversion Completed.\n")

    print("The video file is saved at '{}' named '{}'.".format(destination_path, filename[:-4]+".mp4"))




def img2gif(images_path:str, destination_path:str, filename:str, size:tuple[int,int],
            duration_of_frame:int, loop:int) -> None:

    assert os.path.isdir(images_path), "image_path must be a directory."
    # assert filename[-4:] == ".gif", "filename must endswith '.gif' extensions."

    with iio.get_writer(os.path.join(destination_path,filename[:-4]+".gif"), mode="I", duration=duration_of_frame, loop=loop) as writer:
        
        for file in tqdm.tqdm(natsort.natsorted(os.listdir(images_path))):
            if file.endswith((".jpg",".png",".jpeg")):   
                img = cv2.imread(os.path.join(images_path, file))
                img = cv2.resize(img, size)
                writer.append_data(cv2.cvtColor(img,cv2.COLOR_BGR2RGB))

    print("\nConversion Completed.\n")

    print("The gif file is saved at '{}' named '{}'.".format(destination_path, filename[:-4]+".gif"))